from __future__ import annotations

import logging
import time as time_module
from contextlib import contextmanager
from typing import Dict, List, Optional

import attrs

from hkkang_utils.string import generate_random_str


@attrs.define
class Period:
    start_time: float
    end_time: float
    
    @property
    def elapsed_time(self) -> float:
        return self.end_time - self.start_time

class TimerMetaclass(type):
    def __call__(cls, name: Optional[str] = None):
        if name is None:
            while name is None or name in Timer._cache:
                name = generate_random_str(size=4)
        if name not in Timer._cache.keys():
            obj = cls.__new__(cls)
            obj.__init__()
            Timer._cache[name] = obj
        return Timer._cache[name]

# @attrs.define
class Timer(metaclass=TimerMetaclass):
    _cache: Dict[str, Timer] = dict()
    def __init__(self):
        self.start_time: Optional[float] = None
        self.measured_times: List[Period] = []
        self.paused_times: List[Period] = []
    
    @property
    def name(self) -> str:
        for k, v in Timer._cache.items():
            if v is self:
                return k
        return None
        
    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger(f"Timer.{self.name}")

    @contextmanager
    def pause(self) -> Timer:
        # Pre processing
        current_time = time_module.time()
        self.measured_times.append(Period(self.start_time, current_time))
        self.start_time = current_time
        yield self
        # Post processing
        current_time = time_module.time()
        self.paused_times.append(Period(self.start_time, current_time))
        self.start_time = current_time
    
    @contextmanager
    def measure(self, print_measured_time: bool = False) -> Timer:
        # Pre processing
        self.start_time = time_module.time()
        yield self
        # Post processing
        self.measured_times.append(Period(self.start_time, time_module.time()))
        if print_measured_time:
            self.show_elapsed_time()
        
    @staticmethod
    def get_timer(name: str) -> Timer:
        # __new__ will create a new timer if it does not exist
        return Timer(name)

    @property
    def elapsed_time(self):
        return sum([period.elapsed_time for period in self.measured_times])
    
    def start(self) -> None:
        if self.start_time is not None:
            self.logger.warning("Timer has already been started. Please call stop() first.")
        self.start_time = time_module.time()
        
    def stop(self) -> None:
        if self.start_time is None:
            self.logger.warning("Timer has not been started. Please call start() first.")
        self.measured_times.append(Period(self.start_time, time_module.time()))
        self.start_time = None

    def show_elapsed_time(self):
        self.logger.info(f"Elapsed time: {self.elapsed_time:.2f} seconds")


    def delete(self):
        del self.cache[self.name]


def measure_time(func):
    def wrapper(*args, **kwargs):
        timer = Timer()
        with timer.measure(print_measured_time=True):
            result = func(*args, **kwargs)
        del timer
        return result
    return wrapper

if __name__ == "__main__":
    timer = Timer()
    print("Done")


import functools
import logging
import threading
from concurrent.futures import ProcessPoolExecutor
from typing import Any, Callable, List
import traceback


def _shorten_string(text: str) -> str:
    assert isinstance(text, str), f"text must be str type, but {type(text)} is given."
    if len(text) > 10:
        return text[:10] + "..."
    return text

class Thread(threading.Thread):
    """Please use start method to start thread. start method will call run method.
    If you are doing cpu intensive task, please use MultiProcessor class instead of this class. It's much faster.

    example:
        # Initialize threads
        thread1 = concurrent_utils.Thread(count_million_and_print_name, 1, {"name":"name_1"})
        thread2 = concurrent_utils.Thread(count_million_and_print_name, 2, {"name":"name_2"})

        # Start threads
        thread1.start()
        thread2.start()

        # Wait for threads to finish (this is optional: accessing result will call join method anyway)
        thread1.join()
        thread2.join()

        # Get results
        result1 = thread1.result
        result2 = thread2.result
    """

    # Static variable
    cnt = 0

    def __init__(self, *args, **kwargs):
        super().__init__()
        Thread.cnt += 1
        self.threadID = Thread.cnt
        self.func = args[0]
        self.args = args[1:]
        self.kwargs = kwargs
        self._result = None
        self.logger = logging.getLogger(f"Thread {self.threadID}")

    @property
    def result(self):
        if self.is_alive():
            self.join()
        return self._result

    def run(self):
        post_fix = f"Thread {self.threadID}: {self.func.__name__}(args={_shorten_string(str(self.args))}, kwargs={_shorten_string(str(self.kwargs))})"
        self.logger.info(f"Starting {post_fix}")
        self._result = self.func(*self.args, **self.kwargs)
        self.logger.info(f"Exiting {post_fix}")

    def _parse_args(self, args):
        if args is None:
            return ()
        elif isinstance(args, tuple):
            return args
        elif isinstance(args, list):
            return tuple(args)
        else:
            return (args,)

    def _parse_kwargs(self, kwargs):
        if kwargs is None:
            return {}
        elif isinstance(kwargs, dict):
            return kwargs
        else:
            raise ValueError(f"kwargs must be dict type, but {type(kwargs)} is given.")

class MultiProcessor():
    """
    example:
        # Initialize multi processor
        multiprocessor = Multiprocessor(num_process=4)
        
        # Run processes concurrently
        multiprocessor.run(count_million_and_print_name, 1, name=f"name_{i}")
        multiprocessor.run(count_million_and_print_name, 2, name=f"name_{i}")
        multiprocessor.run(count_million_and_print_name, 3, name=f"name_{i}")
        multiprocessor.run(count_million_and_print_name, 4, name=f"name_{i}")

        # Wait for all processes to finish (this is optional: accessing result will call join method anyway)
        multiprocessor.join()

        # Get results
        results = multiprocessor.results
    """
    def __init__(self, num_workers: int, store_results: bool = True):
        self.num_workers = num_workers
        self._futures = []
        self._results = []
        self.store_results = store_results
        self.logger = logging.getLogger(f"MultiProcessor")

    @property
    def results(self):
        return self._results
    
    @functools.cached_property
    def executor(self):
        return ProcessPoolExecutor(max_workers=self.num_workers)
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, tb):
        # Clean up: Wait for all processes to finish
        self.join()
        
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
            return False
        return True
    
    def _check_if_valid_function(self, func: Callable):
        if not hasattr(func, "__call__"):
            self.logger.error(f"Given function is not callable. Please use callable function.")
        elif func.__name__ == "<lambda>":
            self.logger.error(f"Given function is lambda function. Please use normal function instead of lambda function.")
        else:            
            return True
        return False
    
    def run(self, *args, **kwargs):
        """Pass callable function in the first argument. 
        Then, pass other arguments and keyword arguments as you would to the callable function.
        """
        # Extract function and arguments
        assert len(args) > 0, f"Please pass function as the first argument."
        func: Callable = args[0]
        args = args[1:]
        
        # Check the given function is not lambda function
        self._check_if_valid_function(func)
        
        self._futures.append(self.executor.submit(functools.partial(func, *args, **kwargs)))
    
    def join(self):
        if self.store_results:
            self._results = [future.result() for future in self._futures]
        else:
            for future in self._futures:
                future.result()
        # Free memory
        self.executor.shutdown()

if __name__ == "__main__":
    pass



import unittest

import time as time_module
import src.hkkang_utils.time as time_utils

TEST_TIME = 2

class Test_time_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_time_utils, self).__init__(*args, **kwargs)
        
    @classmethod
    def setUpClass(cls):
        cls.timer = time_utils.Timer()
        with cls.timer.measure():
            time_module.sleep(TEST_TIME)
    
    def test_timer_measure(self):
        timer = time_utils.Timer()
        timer.start()
        time_module.sleep(TEST_TIME)
        timer.stop()
        assert timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME+1, f"Timer is not working properly: {timer.elapsed_time} sec is measured."
    
    def test_timer_measure_as_decorator(self):
        timer = time_utils.Timer()
        with timer.measure():
            time_module.sleep(TEST_TIME)
        assert timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME+1, f"Timer is not working properly: {timer.elapsed_time} sec is measured."
    
    def test_timer_loading_by_name(self):
        timer_name = self.timer.name
        timer = time_utils.Timer(timer_name)
        assert timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME+1, f"Timer is not working properly: {timer.elapsed_time} sec is measured."

if __name__ == "__main__":
    unittest.main()
    
    